;; Generated package description from keycast.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "keycast" "1.4.8.0.20260601.0" "Show current command and its binding" '((emacs "28.1") (compat "31.0") (cond-let "1.1")) :commit "a6518e1b48b08ba883e9b1a2db0872d5bf3d85f4" :authors '(("Jonas Bernoulli" . "emacs.keycast@jonas.bernoulli.dev")) :maintainer '("Jonas Bernoulli" . "emacs.keycast@jonas.bernoulli.dev") :keywords '("multimedia") :url "https://github.com/tarsius/keycast")
